use std::sync::Arc;

use vulkano::{
    Version, VulkanLibrary,
    device::{
        Device, DeviceCreateInfo, DeviceExtensions, DeviceFeatures, Queue, QueueCreateInfo,
        QueueFlags,
        physical::{PhysicalDevice, PhysicalDeviceType},
    },
    instance::{Instance, InstanceCreateFlags, InstanceCreateInfo, InstanceExtensions},
};
use winit::raw_window_handle::HasDisplayHandle;

pub(super) fn load_library() -> Arc<VulkanLibrary> {
    let library = VulkanLibrary::new().unwrap();

    log::debug!("Available layer:");
    for layer in library.layer_properties().unwrap() {
        log::debug!(
            "\t - Layer name: {}, Description: {}, Implementation Version: {}, Vulkan Version: {}",
            layer.name(),
            layer.description(),
            layer.implementation_version(),
            layer.vulkan_version()
        );
    }

    library
}

pub(super) fn create_instance(
    library: Arc<VulkanLibrary>,
    required_extensions: InstanceExtensions,
) -> Arc<Instance> {
    Instance::new(
        library,
        InstanceCreateInfo {
            // Enable enumerating devices that use non-conformant Vulkan implementations.
            // (e.g. MoltenVK)
            flags: InstanceCreateFlags::ENUMERATE_PORTABILITY,
            enabled_extensions: required_extensions,
            enabled_layers: vec![String::from("VK_LAYER_KHRONOS_validation")],
            ..Default::default()
        },
    )
    .unwrap()
}

pub(super) fn find_physical_device_queue_family_indexes(
    physical_device: &Arc<PhysicalDevice>,
    display_handle: &impl HasDisplayHandle,
) -> Option<u32> {
    let mut graphic_queue_family_index = None;

    for (i, queue_family_property) in physical_device.queue_family_properties().iter().enumerate() {
        if queue_family_property
            .queue_flags
            .intersects(QueueFlags::GRAPHICS)
            && physical_device
                .presentation_support(i as u32, display_handle)
                .unwrap()
        {
            graphic_queue_family_index = Some(i as u32);
        }
    }

    graphic_queue_family_index
}

pub(super) fn pick_physical_device_and_queue_family_indexes(
    instance: &Arc<Instance>,
    display_handle: &impl HasDisplayHandle,
    device_extensions: &DeviceExtensions,
) -> Option<(Arc<PhysicalDevice>, u32)> {
    instance
        .enumerate_physical_devices()
        .unwrap()
        .filter(|p| {
            p.api_version() >= Version::V1_3 || p.supported_extensions().khr_dynamic_rendering
        })
        .filter(|p| p.supported_extensions().contains(device_extensions))
        .filter_map(|p| {
            find_physical_device_queue_family_indexes(&p, display_handle)
                .and_then(|indexes| Some((p, indexes)))
        })
        .min_by_key(|(p, _)| match p.properties().device_type {
            PhysicalDeviceType::DiscreteGpu => 0,
            PhysicalDeviceType::IntegratedGpu => 1,
            PhysicalDeviceType::VirtualGpu => 2,
            PhysicalDeviceType::Cpu => 3,
            PhysicalDeviceType::Other => 4,
            _ => 5,
        })
}

pub(super) fn pick_graphics_device(
    instance: &Arc<Instance>,
    display_handle: &impl HasDisplayHandle,
) -> (Arc<Device>, impl ExactSizeIterator<Item = Arc<Queue>>) {
    let mut device_extensions = DeviceExtensions {
        khr_swapchain: true,
        ..DeviceExtensions::empty()
    };

    let (physical_device, graphics_family_index) =
        pick_physical_device_and_queue_family_indexes(instance, display_handle, &device_extensions)
            .unwrap();

    log::debug!(
        "Using device: {} (type: {:?})",
        physical_device.properties().device_name,
        physical_device.properties().device_type,
    );

    if physical_device.api_version() < Version::V1_3 {
        device_extensions.khr_dynamic_rendering = true;
    }

    log::debug!("Using device extensions: {:#?}", device_extensions);

    Device::new(
        physical_device,
        DeviceCreateInfo {
            queue_create_infos: vec![QueueCreateInfo {
                queue_family_index: graphics_family_index,
                ..Default::default()
            }],
            enabled_extensions: device_extensions,
            enabled_features: DeviceFeatures {
                dynamic_rendering: true,
                ..DeviceFeatures::empty()
            },
            ..Default::default()
        },
    )
    .unwrap()
}
